/*
 * File...... main.c
 * Author.... Joe Grand, Grand Idea Studio, Inc. [@joegrand]
 * Updated... 29 Jan 2018
 *
 * This program uses Tomu's (https://tomu.im) red LED as an optical covert channel.
 * Optical covert channels transmit data by modulating visible light in a way that is 
 * undetectable to the human eye. ASCII printable data is sent via the LED with NRZ 
 * (Non-Return-to-Zero) encoding to emulate a standard UART interface.
 *
 * More details about this example and the hardware needed to receive the
 * data can be found at:
 * http://www.grandideastudio.com/portfolio/optical-covert-channels/
 *
 * Code based on the efm32hg-blinky demo: https://github.com/im-tomu/tomu-samples
 *
 */

// Copyright 2016 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <stdint.h>
#include <stdbool.h>

#include "udelay.h"

#include "em_chip.h"
#include "em_cmu.h"
#include "em_device.h"
#include "em_emu.h"
#include "em_gpio.h"

// Constants
#define BIT_TIME_US     52    // Bit time in uS
                              // Baud rate for transmission = 1/BIT_TIME = 19200
#define UDELAY_LFXO     1     // Use LFXO oscillator for delay loop calibration
#define UDELAY_OFFSET   12    // Manually calculated offset for bitbang_char to have the desired timing

char const msg_covert[] = {"I'm Tomu speaking through light waves!\n"};

// Function prototypes
void hardware_init(void);
void gpio_set_value(int);
void bitbang_string(const char*);
void bitbang_char(char);

// Core functions
int main(void)
{
  hardware_init();
  
  while (1)
  {
    bitbang_string(msg_covert);  // Send pre-defined message to LED
    UDELAY_Delay(750);  // Delay to let receiver properly synchronize
    UDELAY_Delay(750);  // Calls to UDELAY_Delay longer than 1000
    UDELAY_Delay(750);  // will give unpredictable results
    UDELAY_Delay(750);  // ...
  }
}

void hardware_init(void)
{
  // Runs the Silicon Labs chip initialization stuff, that also deals with
  // errata (implements workarounds, etc.)
  CHIP_Init();

  // Switch on the clock for GPIO
  CMU_ClockEnable(cmuClock_GPIO, true);
  
  // Set up two pins with the GPIO controller
  // and configure them to be open drain:
  //  - PA0 == green
  //  - PB7 == red
  GPIO_PinModeSet(gpioPortA, 0, gpioModeWiredAnd, 0);
  GPIO_PinModeSet(gpioPortB, 7, gpioModeWiredAnd, 0);

  // Default to both LEDs off
  GPIO_PinOutSet(gpioPortA, 0);
  GPIO_PinOutSet(gpioPortB, 7);
  
  UDELAY_Calibrate(); // Calibrate active-wait microsecond delay
}

void gpio_set_value(int value)
{
    switch (value)
    {
        case 0: // GPIO pin low -> LED on
            GPIO_PinOutClear(gpioPortB, 7);
            break;
        case 1: // GPIO pin high -> LED off
            GPIO_PinOutSet(gpioPortB, 7);
            break;
        default:
            break;
    }
}

void bitbang_string(const char* string)
{
	int i;
 
    for (i = 0; i < (int)strlen(string); ++i)
    {
        bitbang_char(string[i]);
    }
}

void bitbang_char(char character)
{
    int i;
 
    // Start bit
    gpio_set_value(1);
    UDELAY_Delay(BIT_TIME_US - UDELAY_OFFSET);

    // Data bits
    for (i = 0; i < 8; ++i) 
	{ 
        gpio_set_value(!((character >> i) & 1));
        UDELAY_Delay(BIT_TIME_US - UDELAY_OFFSET);
    }
   
    // Stop bit
    gpio_set_value(0);
    UDELAY_Delay(BIT_TIME_US - UDELAY_OFFSET);
}

